@@ -31,23 +31,42 @@ merge.data.table <- function(x, y, by = NULL, all = FALSE, all.x = all,
         length(dt.key) < length(by) || !all(dt.key[1:length(by)] == by)
     }
 
+    ..i = NULL  # to give R CMD check a visible binding
     if (.reset.keys(x, by)) {
-        key(x) <- by
+        # if x has many columns, coping table and setting key on all columns may be relatively slow, so we use a
+        # manual secondary key here. TO DO: replace with set2key when implemented.
+        xkey = x[,by,with=FALSE]  
+        xkey[,..i:=1:nrow(xkey)]
+        setkeyv(xkey,by)
+        xsecondary=TRUE
+    } else {
+        xkey = x   # no copy here
+        xsecondary=FALSE
     }
     if (.reset.keys(y, by)) {
-        key(y) <- by
+        ykey = y[,by,with=FALSE]
+        ykey[,..i:=1:nrow(ykey)]
+        setkeyv(ykey,by)
+        ysecondary=TRUE
+    } else {
+        ykey = y
+        ysecondary=FALSE
     }
-    key <- by
 
-    ykey <- y[,key,with=FALSE]
-    xidx <- x[ykey, nomatch = 0, mult = 'all', which = TRUE]
-    xkey <- x[,key,with=FALSE]
-    yidx <- y[xkey, nomatch = 0, mult = 'all', which = TRUE]
+    xidx = if (xsecondary)
+        xkey[ykey, ..i, nomatch=0, mult="all"]$..i   # TO DO: use drop=TRUE when implemented
+    else
+        x[ykey, nomatch = 0, mult = 'all', which = TRUE]
+        
+    yidx = if (ysecondary)
+        ykey[xkey, ..i, nomatch=0, mult="all"]$..i
+    else
+        y[xkey, nomatch = 0, mult = 'all', which = TRUE]
 
     if (any(xidx)) {
         tmp = sort(xidx)
         xx <- x[tmp]
-        yy <- y[yidx, setdiff(names(y), key), with = FALSE]
+        yy <- y[yidx, setdiff(names(y), by), with = FALSE]
         dt <- xx
         if (ncol(yy) > 0) {
             dt <- cbind(xx, yy)
@@ -60,7 +79,7 @@ merge.data.table <- function(x, y, by = NULL, all = FALSE, all.x = all,
     if (all.x) {
         missingxidx <- setdiff(1:nrow(x), xidx)
         if (length(missingxidx) > 0) {
-            othercols <- setdiff(names(y), key)
+            othercols <- setdiff(names(y), by)
             if (length(othercols) > 0) {
                 tmp = rep(1, length(missingxidx))
                 xx <- cbind(x[missingxidx], y[NA, othercols, with = FALSE][tmp])
@@ -72,20 +91,20 @@ merge.data.table <- function(x, y, by = NULL, all = FALSE, all.x = all,
     if (all.y) {
         missingyidx <- setdiff(1:nrow(y), yidx)
         if (length(missingyidx) > 0) {
-            yy <- y[missingyidx, key, with = FALSE]
-            othercolsx <- setdiff(names(x), key)
+            yy <- y[missingyidx, by, with = FALSE]
+            othercolsx <- setdiff(names(x), by)
             if (length(othercolsx) > 0) {
                 tmp = rep(1, length(missingyidx))
                 yy <- cbind(yy, x[NA, othercolsx, with = FALSE][tmp])
             }
-            othercolsy <- setdiff(names(y), key)
+            othercolsy <- setdiff(names(y), by)
             if (length(othercolsy) > 0)
                 yy <- cbind(yy, y[missingyidx, othercolsy, with = FALSE])
             dt <- rbind(dt, yy[,names(dt),with=FALSE])
         }
     }
-    if (nrow(dt) > 0) key(dt) = key
-
+    if (nrow(dt) > 0) setkeyv(dt,by)
+    
     ###########################################################################
     ## Handle merging suffix behavior.
     ## Enable ability to use "old" suffix behavior for now.
@@ -135,7 +154,6 @@ merge.data.table <- function(x, y, by = NULL, all = FALSE, all.x = all,
             }
         }
     }
-
     dt
 }
 
