@@ -1,5 +1,5 @@
 
-test.data.table = function() {
+test.data.table = function(echo=FALSE) {
     if (exists("test.data.table",.GlobalEnv,inherits=FALSE)) {
         # package developer
         if ("package:data.table" %in% search()) stop("data.table package loaded")
@@ -12,7 +12,10 @@ test.data.table = function() {
     # for (fn in dir(d,"*.[rR]$",full=TRUE)) {  # testthat runs those
     for (fn in file.path(d, 'tests.Rraw')) {    # not testthat
         cat("Running",fn,"\n")
-        sys.source(fn,envir=new.env(parent=.GlobalEnv))
+        oldverbose = getOption("datatable.verbose",FALSE)
+        if (echo) options(datatable.verbose=TRUE)
+        sys.source(fn,envir=new.env(parent=.GlobalEnv))  # using source(local=) would break 2.12 compatibility
+        options(data.table.verbose=oldverbose)
         # the new.env() is required for when a *user* runs test.data.table() because
         # the context of this function is sealed in the namespace w.r.t S4.
         # It is also tidier to protect the tests from the variable 'd' above.
@@ -33,29 +36,35 @@ test.data.table = function() {
 ## built around test_that. A call to test_that::{expect_equal|equal} will
 ## ultimately dispatch to this method when making an "equality" call.
 all.equal.data.table <- function(target, current, trim.levels=TRUE, ...) {
-    force(target)
-    force(current)
+    target = copy(target)
+    current = copy(current)
     if (trim.levels) {
         ## drop unused levels
         if (length(target)) {
             for (i in which(sapply(target, is.factor))) {
-                target[[i]] <- factor(target[[i]])
+                .xi = factor(target[[i]])
+                target[,i:=.xi,with=FALSE]
             }
         }
         if (length(current)) {
             for (i in which(sapply(current, is.factor))) {
-                current[[i]] <- factor(current[[i]])
+                .xi = factor(current[[i]])
+                current[,i:=.xi,with=FALSE]
             }
         }
     }
 
     ## Trim any extra row.names attributes that came from some inheritence
-    if (length(attr(target, "row.names"))) {
-        setattr(target, "row.names", NULL)
-    }
-    if (length(attr(current, "row.names"))) {
-        setattr(current, "row.names", NULL)
-    }
+    setattr(target, "row.names", NULL)
+    setattr(current, "row.names", NULL)
+    
+    # all.equal uses unclass which doesn't know about external pointers : there
+    # doesn't seem to be all.equal.externalptr method in base.
+    setattr(target, ".internal.selfref", NULL)
+    setattr(current, ".internal.selfref", NULL)
     
     all.equal.list(target, current, ...)
 }
+
+
+
